

class CompsRecordException(Exception):

    def __init__(self, record_number, message):
        super().__init__(message + f"Unknown {message} version found- {record_number}")


class UnknownRxTagVersion(CompsRecordException):

    def __init__(self, record_number):
        super().__init__(record_number, "RxTag")



import os
import shutil
import struct
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from sqlite3 import Cursor
from typing import List, Tuple, Dict

from acd.exceptions.CompsRecordException import UnknownRxTagVersion
from acd.generated.comps.rx_tag import RxTag
from acd.generated.controller.rx_controller import RxController
from acd.generated.map_device.rx_map_device import RxMapDevice


@dataclass
class L5xElementBuilder:
    _cur: Cursor
    _object_id: int = -1


@dataclass
class L5xElement:
    name: str


@dataclass
class DataType(L5xElement):
    children: List[str]


@dataclass
class Tag(L5xElement):
    data_table_instance: int
    data_type: str
    comments: List[Tuple[str, str]]


@dataclass
class MapDevice(L5xElement):
    module_id: int
    parent_module: int
    slot_no: int
    vendor_id: int
    product_type: int
    product_code: int
    comments: List[Tuple[str, str]]


@dataclass
class Routine(L5xElement):
    rungs: List[str]


@dataclass
class AOI(L5xElement):
    routines: List[Routine]
    tags: List[Tag]


@dataclass
class Program(L5xElement):
    routines: List[Routine]
    tags: List[Tag]


@dataclass
class Controller(L5xElement):
    serial_number: str
    path: str
    data_types: List[DataType]
    tags: List[Tag]
    programs: List[Program]
    aois: List[AOI]
    map_devices: List[MapDevice]


@dataclass
class DataTypeBuilder(L5xElementBuilder):

    def build(self) -> DataType:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id))
        member_results = self._cur.fetchall()
        if len(member_results) == 1:
            member_collection_id = member_results[0][1]

            self._cur.execute(
                "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                    member_collection_id))
            children_results = self._cur.fetchall()
            children = []
            for child in children_results:
                children.append(child[0])
            return DataType(name, children)
        return DataType(name, [])


@dataclass
class MapDeviceBuilder(L5xElementBuilder):

    def build(self) -> MapDevice:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        r = RxMapDevice.from_bytes(results[0][3])

        if r.record_format_version == 0x00:
            return MapDevice(results[0][0], 0, 0, 0, 0, 0, 0, [])
        elif not r.body.valid:
            return MapDevice("", 0, 0, 0, 0, 0, 0, [])

        self._cur.execute(
            "SELECT tag_reference, record_string FROM comments WHERE parent=" + str(
                r.comment_id))
        comment_results = self._cur.fetchall()

        name = results[0][0]
        return MapDevice(name, r.body.module_id, r.body.parent_module, r.body.slot_no, r.body.vendor_id, r.body.product_type, r.body.product_code, comment_results)


@dataclass
class TagBuilder(L5xElementBuilder):

    def build(self) -> Tag:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        r = RxTag.from_bytes(results[0][3])

        if r.record_format_version == 0x00:
            return Tag(results[0][0], 0, "", [])
        elif not r.body.valid:
            raise UnknownRxTagVersion(r.record_format_version)

        if r.body.data_type == 4294967295:
            data_type = ""
            name = r.body.name
            comment_results = []
        else:
            self._cur.execute(
                "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                    r.body.data_type))
            data_type_results = self._cur.fetchall()
            data_type = data_type_results[0][0]

            self._cur.execute(
                "SELECT tag_reference, record_string FROM comments WHERE parent=" + str(
                    r.comment_id))
            comment_results = self._cur.fetchall()
            try:
                name = r.body.name
            except Exception as e:
                name = ""
            if len(comment_results) > 0:
                pass
        if r.body.dimension_1 != 0:
            data_type = data_type + "[" + str(r.body.dimension_1) + "]"
        if r.body.dimension_2 != 0:
            data_type = data_type + "[" + str(r.body.dimension_2) + "]"
        if r.body.dimension_3 != 0:
            data_type = data_type + "[" + str(r.body.dimension_3) + "]"
        return Tag(name, r.body.data_table_instance, data_type, comment_results)


@dataclass
class RoutineBuilder(L5xElementBuilder):

    def build(self) -> Routine:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT object_id, parent_id, seq_no FROM region_map WHERE parent_id=" + str(
                self._object_id) +  " ORDER BY seq_no")
        results = self._cur.fetchall()
        rungs = []
        for member in results:
            self._cur.execute(
                "SELECT object_id, rung FROM rungs WHERE object_id=" + str(
                    member[0]))
            rungs_results = self._cur.fetchall()
            if len(rungs_results) > 0:
                rungs.append(rungs_results[0][1])
        return Routine(name, rungs)


@dataclass
class AoiBuilder(L5xElementBuilder):

    def build(self) -> AOI:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxRoutineCollection'")
        collection_results = self._cur.fetchall()
        if len(collection_results) != 0:
            collection_id = collection_results[0][1]
        else:
            routines = []
            tags: List[Tag] = []
            return AOI(name, routines, tags)

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                collection_id))
        routine_results = self._cur.fetchall()

        routines = []
        for child in routine_results:
            routines.append(RoutineBuilder(self._cur, child[1]).build())

        # Get the Program Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one program tag collection")

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                results[0][1]))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            tags.append(TagBuilder(self._cur, result[1]).build())

        return AOI(name, routines, tags)


@dataclass
class ProgramBuilder(L5xElementBuilder):

    def build(self) -> Program:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxRoutineCollection'")
        collection_results = self._cur.fetchall()
        collection_id = collection_results[0][1]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                collection_id))
        routine_results = self._cur.fetchall()

        routines = []
        for child in routine_results:
            routines.append(RoutineBuilder(self._cur, child[1]).build())

        # Get the Program Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one program tag collection")

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                results[0][1]))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            tags.append(TagBuilder(self._cur, result[1]).build())

        return Program(name, routines, tags)


@dataclass
class ControllerBuilder(L5xElementBuilder):

    def build(self) -> Controller:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type, record FROM comps WHERE parent_id=0 AND record_type=256")
        results = self._cur.fetchall()
        if len(results) != 1:
            raise Exception("Does not contain exactly one root controller node")
        serial_number = ""
        path = ""
        try:
            r = RxController.from_bytes(results[0][4])
            serial_number = hex(r.body.serial_number)
            if r.record_format_version == 103:
                path = r.body.path
        except:
            pass

        self._object_id = results[0][1]
        controller_name = results[0][0]

        # Get the data types
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxDataTypeCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller data type collection")

        _data_type_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _data_type_id))
        results = self._cur.fetchall()

        data_types: List[DataType] = []
        for result in results:
            _data_type_object_id = result[1]
            data_types.append(DataTypeBuilder(self._cur, _data_type_object_id).build())

        # Get the Controller Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller tag collection")
        _tag_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _tag_collection_object_id))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            _tag_object_id = result[1]
            tags.append(TagBuilder(self._cur, _tag_object_id).build())

        # Get the Program Collection and get the programs
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxProgramCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller program collection")

        _program_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _program_collection_object_id))
        results = self._cur.fetchall()
        programs: List[Program] = []
        for result in results:
            _program_object_id = result[1]
            programs.append(ProgramBuilder(self._cur, _program_object_id).build())

        # Get the AOI Collection and get the AOIs
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxUDIDefinitionCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one AOI collection")
        _aoi_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _aoi_collection_object_id))
        results = self._cur.fetchall()
        aois: List[AOI] = []
        for result in results:
            _aoi_object_id = result[1]
            aois.append(AoiBuilder(self._cur, _aoi_object_id).build())

        # Get the Map Device (IO) Collection and get the MapDevices
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxMapDeviceCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one Map Device collection")
        _map_device_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _map_device_collection_object_id))
        results = self._cur.fetchall()
        map_devices: List[MapDevice] = []
        for result in results:
            _map_device_object_id = result[1]
            map_devices.append(MapDeviceBuilder(self._cur, _map_device_object_id).build())

        return Controller(controller_name, serial_number, path, data_types, tags, programs, aois, map_devices)

@dataclass
class DumpCompsRecords(L5xElementBuilder):
    base_directory: PathLike = Path("dump")

    def dump(self, parent_id: int = 0):
        self._cur.execute(
            f"SELECT comp_name, object_id, parent_id, record_type, record FROM comps WHERE parent_id={parent_id}")
        results = self._cur.fetchall()

        for result in results:
            object_id = result[1]
            name = result[0]
            record = result[4]
            new_path = Path(os.path.join(self.base_directory, name))
            if os.path.exists(os.path.join(new_path)):
                shutil.rmtree(os.path.join(new_path))
            if not os.path.exists(os.path.join(new_path)):
                os.makedirs(new_path)
            with open(Path(os.path.join(new_path, name + ".dat")), "wb") as file:
                file.write(record)

            DumpCompsRecords(self._cur, object_id, new_path).dump(object_id)





import os
from dataclasses import dataclass
from loguru import logger as log


from acd.dbextract import DbExtract
from acd.unzip import Unzip


@dataclass
class AcdDatabase:
    input_filename: os.PathLike
    output_filename: str
    _temp_dir: str = "build"  # tempfile.mkdtemp()

    def __post_init__(self):
        if not os.path.exists(os.path.join(self._temp_dir)):
            os.makedirs(self._temp_dir)

        log.info("Extracting ACD database file")
        unzip = Unzip(self.input_filename)
        unzip.write_files(self._temp_dir)

        log.info("Getting records from ACD Comps file and storing in sqllite database")
        self.comps_db = DbExtract(os.path.join(self._temp_dir, "Comps.Dat")).read()

        log.info("Getting records from ACD SbRegion file and storing in sqllite database")
        self.sb_region_db = DbExtract(os.path.join(self._temp_dir, "SbRegion.Dat")).read()

        log.info("Getting records from ACD Comments file and storing in sqllite database")
        self.comments_db = DbExtract(os.path.join(self._temp_dir, "Comments.Dat")).read()

        log.info("Getting records from ACD Nameless file and storing in sqllite database")
        self.nameless_db = DbExtract(os.path.join(self._temp_dir, "Nameless.Dat")).read()

    def extract_to_file(self):
        directory = os.path.join(self._temp_dir, "comps_db")
        if not os.path.exists(os.path.join(directory)):
            os.makedirs(directory)
        for count, record in enumerate(self.comps_db.records.record):
            with open(os.path.join(directory, str(count)), "wb") as out_file:
                out_file.write(record.record.record_buffer)

        directory = os.path.join(self._temp_dir, "sb_region_db")
        if not os.path.exists(os.path.join(directory)):
            os.makedirs(directory)
        for count, record in enumerate(self.sb_region_db.records.record):
            with open(os.path.join(directory, str(count)), "wb") as out_file:
                out_file.write(record.record.record_buffer)

        directory = os.path.join(self._temp_dir, "comments_db")
        if not os.path.exists(os.path.join(directory)):
            os.makedirs(directory)
        for count, record in enumerate(self.comments_db.records.record):
            with open(os.path.join(directory, str(count)), "wb") as out_file:
                out_file.write(record.record.record_buffer)

        directory = os.path.join(self._temp_dir, "nameless_db")
        if not os.path.exists(os.path.join(directory)):
            os.makedirs(directory)
        for count, record in enumerate(self.nameless_db.records.record):
            with open(os.path.join(directory, str(count)), "wb") as out_file:
                out_file.write(record.record.record_buffer)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class RxController(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.parent_id = self._io.read_u4le()
        self.unique_tag_identifier = self._io.read_u4le()
        self.record_format_version = self._io.read_u2le()
        self.comment_id = self._io.read_u4le()
        _on = self.record_format_version
        if _on == 95:
            self.body = RxController.V95(self._io, self, self._root)
        elif _on == 103:
            self.body = RxController.V103(self._io, self, self._root)
        else:
            self.body = RxController.VUnknown(self._io, self, self._root)

    class VUnknown(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = False
            return getattr(self, '_m_valid', None)


    class V95(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def len_most_recent(self):
            if hasattr(self, '_m_len_most_recent'):
                return self._m_len_most_recent

            _pos = self._io.pos()
            self._io.seek(363)
            self._m_len_most_recent = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_most_recent', None)

        @property
        def record(self):
            if hasattr(self, '_m_record'):
                return self._m_record

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_record = self._io.read_bytes(self.len_record)
            self._io.seek(_pos)
            return getattr(self, '_m_record', None)

        @property
        def len_current_active(self):
            if hasattr(self, '_m_len_current_active'):
                return self._m_len_current_active

            _pos = self._io.pos()
            self._io.seek(327)
            self._m_len_current_active = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_current_active', None)

        @property
        def most_recent(self):
            if hasattr(self, '_m_most_recent'):
                return self._m_most_recent

            _pos = self._io.pos()
            self._io.seek(367)
            self._m_most_recent = (self._io.read_bytes(self.len_most_recent)).decode(u"utf-16")
            self._io.seek(_pos)
            return getattr(self, '_m_most_recent', None)

        @property
        def serial_number(self):
            if hasattr(self, '_m_serial_number'):
                return self._m_serial_number

            _pos = self._io.pos()
            self._io.seek(459)
            self._m_serial_number = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_serial_number', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def len_record(self):
            if hasattr(self, '_m_len_record'):
                return self._m_len_record

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_len_record = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_record', None)

        @property
        def current_acive(self):
            if hasattr(self, '_m_current_acive'):
                return self._m_current_acive

            _pos = self._io.pos()
            self._io.seek(331)
            self._m_current_acive = (self._io.read_bytes(self.len_current_active)).decode(u"utf-16")
            self._io.seek(_pos)
            return getattr(self, '_m_current_acive', None)


    class V103(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def len_most_recent(self):
            if hasattr(self, '_m_len_most_recent'):
                return self._m_len_most_recent

            _pos = self._io.pos()
            self._io.seek(232)
            self._m_len_most_recent = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_most_recent', None)

        @property
        def record(self):
            if hasattr(self, '_m_record'):
                return self._m_record

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_record = self._io.read_bytes(self.len_record)
            self._io.seek(_pos)
            return getattr(self, '_m_record', None)

        @property
        def len_current_active(self):
            if hasattr(self, '_m_len_current_active'):
                return self._m_len_current_active

            _pos = self._io.pos()
            self._io.seek(196)
            self._m_len_current_active = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_current_active', None)

        @property
        def most_recent(self):
            if hasattr(self, '_m_most_recent'):
                return self._m_most_recent

            _pos = self._io.pos()
            self._io.seek(236)
            self._m_most_recent = (self._io.read_bytes(self.len_most_recent)).decode(u"utf-16")
            self._io.seek(_pos)
            return getattr(self, '_m_most_recent', None)

        @property
        def serial_number(self):
            if hasattr(self, '_m_serial_number'):
                return self._m_serial_number

            _pos = self._io.pos()
            self._io.seek(328)
            self._m_serial_number = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_serial_number', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def len_record(self):
            if hasattr(self, '_m_len_record'):
                return self._m_len_record

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_len_record = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_record', None)

        @property
        def path(self):
            if hasattr(self, '_m_path'):
                return self._m_path

            _pos = self._io.pos()
            self._io.seek(388)
            self._m_path = (self._io.read_bytes(self.len_path)).decode(u"utf-16")
            self._io.seek(_pos)
            return getattr(self, '_m_path', None)

        @property
        def len_path(self):
            if hasattr(self, '_m_len_path'):
                return self._m_len_path

            _pos = self._io.pos()
            self._io.seek(384)
            self._m_len_path = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_len_path', None)

        @property
        def current_acive(self):
            if hasattr(self, '_m_current_acive'):
                return self._m_current_acive

            _pos = self._io.pos()
            self._io.seek(200)
            self._m_current_acive = (self._io.read_bytes(self.len_current_active)).decode(u"utf-16")
            self._io.seek(_pos)
            return getattr(self, '_m_current_acive', None)






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class RxMapDevice(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.parent_id = self._io.read_u4le()
        self.unique_tag_identifier = self._io.read_u4le()
        self.record_format_version = self._io.read_u2le()
        self.comment_id = self._io.read_u4le()
        _on = self.record_format_version
        if _on == 0:
            self.body = RxMapDevice.V0(self._io, self, self._root)
        elif _on == 162:
            self.body = RxMapDevice.V162(self._io, self, self._root)
        elif _on == 173:
            self.body = RxMapDevice.V173(self._io, self, self._root)
        else:
            self.body = RxMapDevice.VUnknown(self._io, self, self._root)

    class VUnknown(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = False
            return getattr(self, '_m_valid', None)


    class V0(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = False
            return getattr(self, '_m_valid', None)


    class V162(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def module_id(self):
            if hasattr(self, '_m_module_id'):
                return self._m_module_id

            _pos = self._io.pos()
            self._io.seek(126)
            self._m_module_id = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_module_id', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def record_length(self):
            if hasattr(self, '_m_record_length'):
                return self._m_record_length

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_record_length = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_record_length', None)

        @property
        def product_type(self):
            if hasattr(self, '_m_product_type'):
                return self._m_product_type

            _pos = self._io.pos()
            self._io.seek(94)
            self._m_product_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_product_type', None)

        @property
        def vendor_id(self):
            if hasattr(self, '_m_vendor_id'):
                return self._m_vendor_id

            _pos = self._io.pos()
            self._io.seek(92)
            self._m_vendor_id = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_vendor_id', None)

        @property
        def slot_no(self):
            if hasattr(self, '_m_slot_no'):
                return self._m_slot_no

            _pos = self._io.pos()
            self._io.seek(122)
            self._m_slot_no = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_slot_no', None)

        @property
        def product_code(self):
            if hasattr(self, '_m_product_code'):
                return self._m_product_code

            _pos = self._io.pos()
            self._io.seek(96)
            self._m_product_code = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_product_code', None)

        @property
        def parent_module(self):
            if hasattr(self, '_m_parent_module'):
                return self._m_parent_module

            _pos = self._io.pos()
            self._io.seek(112)
            self._m_parent_module = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_parent_module', None)


    class V173(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def module_id(self):
            if hasattr(self, '_m_module_id'):
                return self._m_module_id

            _pos = self._io.pos()
            self._io.seek((((80 + 4) + (self.record_count * 12)) + 38))
            self._m_module_id = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_module_id', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def record_count(self):
            if hasattr(self, '_m_record_count'):
                return self._m_record_count

            _pos = self._io.pos()
            self._io.seek(78)
            self._m_record_count = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_record_count', None)

        @property
        def record_length(self):
            if hasattr(self, '_m_record_length'):
                return self._m_record_length

            _pos = self._io.pos()
            self._io.seek(74)
            self._m_record_length = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_record_length', None)

        @property
        def product_type(self):
            if hasattr(self, '_m_product_type'):
                return self._m_product_type

            _pos = self._io.pos()
            self._io.seek(((80 + 2) + (self.record_count * 12)))
            self._m_product_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_product_type', None)

        @property
        def vendor_id(self):
            if hasattr(self, '_m_vendor_id'):
                return self._m_vendor_id

            _pos = self._io.pos()
            self._io.seek((80 + (self.record_count * 12)))
            self._m_vendor_id = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_vendor_id', None)

        @property
        def slot_no(self):
            if hasattr(self, '_m_slot_no'):
                return self._m_slot_no

            _pos = self._io.pos()
            self._io.seek((((80 + 4) + (self.record_count * 12)) + 22))
            self._m_slot_no = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_slot_no', None)

        @property
        def product_code(self):
            if hasattr(self, '_m_product_code'):
                return self._m_product_code

            _pos = self._io.pos()
            self._io.seek(((80 + 4) + (self.record_count * 12)))
            self._m_product_code = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_product_code', None)

        @property
        def records(self):
            if hasattr(self, '_m_records'):
                return self._m_records

            _pos = self._io.pos()
            self._io.seek(80)
            self._m_records = []
            for i in range(self.record_count):
                self._m_records.append(self._io.read_bytes(12))

            self._io.seek(_pos)
            return getattr(self, '_m_records', None)

        @property
        def parent_module(self):
            if hasattr(self, '_m_parent_module'):
                return self._m_parent_module

            _pos = self._io.pos()
            self._io.seek(((80 + (self.record_count * 12)) + 20))
            self._m_parent_module = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_parent_module', None)






import os
import sqlite3
import tempfile
from dataclasses import dataclass
from sqlite3 import Cursor, Connection

from loguru import logger as log

class DatabaseProvider:
    pass


@dataclass
class SqlDatabaseProvider(DatabaseProvider):
    directory: str = tempfile.mkdtemp()
    filename: str = "acd.db"
    _db: Connection = None
    _cur: Cursor = None

    def __post_init__(self):
        log.info("Creating temporary directory (if it doesn't exist to store ACD database files - " + self.directory)

        if os.path.exists(os.path.join(self.directory, self.filename)):
            os.remove(os.path.join(self.directory, self.filename))
        if not os.path.exists(os.path.join(self.directory)):
            os.makedirs(self.directory)
        log.info("Creating sqllite database to store ACD database records")
        self._db = sqlite3.connect(os.path.join(self.directory, self.filename))
        self._cur: Cursor = self._db.cursor()

        log.debug("Create Comps table in sqllite db")
        self._cur.execute("CREATE TABLE comps(object_id int, parent_id int, comp_name text, seq_number int, "
                          "record_type int, record BLOB NOT NULL)")

        log.debug("Create pointers table in sqllite db")
        self._cur.execute("CREATE TABLE pointers(object_id int, parent_id int, comp_name text, seq_number int, "
                          "record_type int, record BLOB NOT NULL)")

        log.debug("Create Rungs table in sqllite db")
        self._cur.execute("CREATE TABLE rungs(object_id int, rung text, seq_number int)")

        log.debug("Create Region_map table in sqllite db")
        self._cur.execute("CREATE TABLE region_map(object_id int, parent_id int, unknown int, seq_no int, record BLOB "
                          "NOT NULL)")
        log.debug("Create Comments table in sqllite db")
        self._cur.execute(
            "CREATE TABLE comments(seq_number int, string_length int, lookup_id int, comment text, record_type int, "
            "sub_record_type int)")

        log.debug("Create Nameless table in sqllite db")
        self._cur.execute(
            "CREATE TABLE nameless(object_id int, parent_id int, record BLOB NOT NULL)")


# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class FafaComents(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.record_length = self._io.read_u4le()
        self._raw_header = self._io.read_bytes(10)
        _io__raw_header = KaitaiStream(BytesIO(self._raw_header))
        self.header = FafaComents.Header(_io__raw_header, self, self._root)
        _on = self.header.record_type
        if _on == 14:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.Utf16Record(12, _io__raw_body, self, self._root)
        elif _on == 4:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.Utf16Record(12, _io__raw_body, self, self._root)
        elif _on == 1:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.AsciiRecord(_io__raw_body, self, self._root)
        elif _on == 13:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.Utf16Record(12, _io__raw_body, self, self._root)
        elif _on == 3:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.Utf16Record(12, _io__raw_body, self, self._root)
        elif _on == 23:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.ControllerRecord(_io__raw_body, self, self._root)
        elif _on == 2:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.AsciiRecord(_io__raw_body, self, self._root)
        elif _on == 25:
            self._raw_body = self._io.read_bytes((self.record_length - 10))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = FafaComents.ControllerRecord(_io__raw_body, self, self._root)
        else:
            self.body = self._io.read_bytes((self.record_length - 10))

    class ControllerRecord(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown_1 = self._io.read_bytes(8)
            self.object_id = self._io.read_u4le()
            self.unknown_2 = self._io.read_bytes(4)
            self.tag_reference = FafaComents.StrzUtf16(self._io, self, self._root)
            self.unknown_3 = self._io.read_bytes(12)
            self.record_string = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")


    class AsciiRecord(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown_1 = self._io.read_bytes(13)
            self.object_id = self._io.read_u4le()
            self.unknown_2 = self._io.read_bytes(13)
            self.record_string = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")


    class AsciiRecord4(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown_1 = self._io.read_bytes(8)
            self.object_id = self._io.read_u4le()
            self.unknown_2 = self._io.read_bytes(24)
            self.record_string = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")


    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def seq_number(self):
            if hasattr(self, '_m_seq_number'):
                return self._m_seq_number

            _pos = self._io.pos()
            self._io.seek(0)
            self._m_seq_number = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_seq_number', None)

        @property
        def record_type(self):
            if hasattr(self, '_m_record_type'):
                return self._m_record_type

            _pos = self._io.pos()
            self._io.seek(2)
            self._m_record_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_record_type', None)

        @property
        def sub_record_length(self):
            if hasattr(self, '_m_sub_record_length'):
                return self._m_sub_record_length

            _pos = self._io.pos()
            self._io.seek(4)
            self._m_sub_record_length = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_sub_record_length', None)

        @property
        def parent(self):
            if hasattr(self, '_m_parent'):
                return self._m_parent

            _pos = self._io.pos()
            self._io.seek(6)
            self._m_parent = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_parent', None)


    class StrzUtf16(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = (self._io.read_bytes((2 * (len(self.code_units) - 1)))).decode(u"utf-16le")
            self.term = self._io.read_u2le()
            if not self.term == 0:
                raise kaitaistruct.ValidationNotEqualError(0, self.term, self._io, u"/types/strz_utf_16/seq/1")

        @property
        def code_units(self):
            if hasattr(self, '_m_code_units'):
                return self._m_code_units

            _pos = self._io.pos()
            self._io.seek(self._io.pos())
            self._m_code_units = []
            i = 0
            while True:
                _ = self._io.read_u2le()
                self._m_code_units.append(_)
                if _ == 0:
                    break
                i += 1
            self._io.seek(_pos)
            return getattr(self, '_m_code_units', None)


    class Utf16Record(KaitaiStruct):
        def __init__(self, zero_buffer_length, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.zero_buffer_length = zero_buffer_length
            self._read()

        def _read(self):
            self.unknown_1 = self._io.read_bytes(8)
            self.object_id = self._io.read_u4le()
            self.unknown_2 = self._io.read_bytes(4)
            self.len_record = self._io.read_u2le()
            self.tag_reference = FafaComents.StrzUtf16(self._io, self, self._root)
            self.unknown_3 = self._io.read_bytes(self.zero_buffer_length)
            self.record_string = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")


    @property
    def lookup_id(self):
        if hasattr(self, '_m_lookup_id'):
            return self._m_lookup_id

        _pos = self._io.pos()
        self._io.seek(27)
        self._m_lookup_id = self._io.read_u2le()
        self._io.seek(_pos)
        return getattr(self, '_m_lookup_id', None)

    @property
    def sub_record_type(self):
        if hasattr(self, '_m_sub_record_type'):
            return self._m_sub_record_type

        _pos = self._io.pos()
        self._io.seek(41)
        self._m_sub_record_type = self._io.read_u2le()
        self._io.seek(_pos)
        return getattr(self, '_m_sub_record_type', None)


